
# Copyright 2012 Dian-Je Tsai and Wantoto Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from fabric.api import env
from six import callable

default_settings_file_dir = os.path.dirname(__file__)

# Script setup
VERBOSE = 0
CONFIRM_BEFORE_DEPLOY = True

# Connection settings
USE_SSH_CONFIG = False
WEB_SERVERS = []

# Local source code
SOURCE_ROOT = lambda: os.getcwd()
IGNORE_FILE_LIST = lambda: os.path.join(callable(SOURCE_ROOT) and SOURCE_ROOT() or SOURCE_ROOT, '.gitignore')
PROJECT_SETTINGS = lambda: '%s.settings' % env.project_name
PRODUCTION_SETTINGS = lambda: '%s.production_settings' % env.project_name
CHECK_COMMENTS = True
CHECK_TODO = False
CHECK_NOTE = False
CHECK_FIX_ME = True
COLLECT_STATIC = True
#USE_SYMBOLIC_LINK_SETTINGS =  os.path.islink(project_settings_path

# Remote base
REMOTE_PROJECT_NAME = lambda: env.project_name
RUN_REMOTE_SETUP = False
REMOTE_BASE = '/var/django/apps'
ADDITIONAL_IGNORE_LIST = []
CLEAN_PY_IN_REMOTE = False
CHECK_PIP_REQUIREMENTS = True
COMPILE_PYC = True
REMOTE_ENVIRONMENTS = {}
UPDATE_YUM_PACKAGES = False
YUM_PACKAGES = []

# Apache
APACHE_CONF_FILE = lambda: env.project_name
APACHE_CONF_DIR = '/etc/httpd/conf.d'
APACHE_CONF_TEMPLATE_FILENAME = lambda: 'apache.conf.jinja2'
APACHE_CONF_TEMPLATE_PATH = lambda: os.path.join(
    default_settings_file_dir,
    callable(APACHE_CONF_TEMPLATE_FILENAME) and APACHE_CONF_TEMPLATE_FILENAME() or APACHE_CONF_TEMPLATE_FILENAME
)
RESTART_APACHE = False
INSTALL_APACHE_CONF = True
UPDATE_APACHE_CONF = False

# MOD_WSGI
AWS_S3_BASE_URL = None
MEDIA_URL = '/media/'
REDIRECT_STATIC_TO_S3 = False
STATIC_URL = '/static/'
SERVE_FAVICON = False
SERVE_MEDIA = False
SERVE_ROBOTS = False
SERVE_STATIC = True
URL_PREFIX = '/'
WSGI_SCRIPT_DIR_NAME = lambda: env.project_name
WSGI_SCRIPT_FILE_NAME = 'wsgi.py'
WSGI_PROCESS_GROUP = 'apache'
WSGI_PROCESS_NAME = lambda: env.project_name
WSGI_PROCESS_USER = 'apache'
WSGI_PROCESSES_COUNT = 2
WSGI_PYTHON_OPTIMIZE_LEVEL = 1
WSGI_THREADS_COUNT = 15
#MEDIA_PATH = os.path.join(remote_project_source_root, 'media')
#STATIC_PATH = os.path.join(remote_project_source_root, 'static')
#VIRTUALENV_SITE_PACKAGES_PATH = os.path.join(remote_project_venv, 'lib/python%s/site-packages')  # Python ver



"""
Copyright 2012 Dian-Je Tsai and Wantoto Inc

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from functools import update_wrapper
import httplib
from django import VERSION as DJANGO_VERSION
from django.conf import settings
from django.http import HttpResponse, QueryDict
from django.utils.decorators import classonlymethod
from eggplant.constants import HTTP_METHOD
from eggplant.constants.settings import HTTPS_ONLY
from eggplant.views.exceptions import ViewError, MethodNotAllowedError

class View(object):
    def __init__(self, *args, **kwargs):
        self.init_args = args
        self.init_kwargs = kwargs

    @classonlymethod
    def view(cls, *init_args, **init_kwargs):
        def view(request, *args, **kwargs):
            self = cls(*init_args, **init_kwargs)

            # Automatically add 'head' method
            if hasattr(self, 'get') and not hasattr(self, 'head'):
                def _head(self, *args, **kwargs):
                    getattr(self, 'get')(*args, **kwargs)
                    self.response.content = ''
                setattr(self, 'head', _head)

            return self.handle(request, *args, **kwargs)

        # take name and docstring from class
        update_wrapper(view, cls, updated=())

        # and possible attributes set by decorators
        # like csrf_exempt from dispatch
        update_wrapper(view, cls.handle, assigned=())
        return view

    def handle(self, request, *args, **kwargs):
        self.request = request
        self.response = HttpResponse()

        # Go
        try:
            # HTTPS?
            if self.only_accept_https(*args, **kwargs) and not self.request.is_secure():
                raise ViewError(httplib.BAD_REQUEST, 'This url accepts HTTPS only.')

            # Check allowed http methods
            if request.method.upper() in HTTP_METHOD and \
               hasattr(self, request.method.lower()):
                handler = getattr(self, request.method.lower())
            else:
                raise MethodNotAllowedError(request)

            # Check HTTP Refer
            http_referer = self.request.META.get('HTTP_REFERER', None)
            self.check_http_referer(http_referer, *args, **kwargs)

            # Authenticate
            self.authenticate(*args, **kwargs)

            # Dispatch
            self.pre_handle(*args, **kwargs)
            handler(*args, **kwargs)
            self.post_handle(*args, **kwargs)

        # Handler
        except ViewError, e:
            self.response.content = ''
            self.response.status_code = e.status_code

            if e.error_handler is not None:
                error_handler = e.error_handler
            else:
                error_handler = self.default_error_handler
            error_handler(e)

        return self.response

    def pre_handle(self, *args, **kwargs):
        pass

    def post_handle(self, *args, **kwargs):
        pass

    def only_accept_https(self, *args, **kwargs):
        return getattr(settings, HTTPS_ONLY, False)

    def authenticate(self, *args, **kwargs):
        pass

    def check_http_referer(self, referer, *args, **kwargs):
        #raise exceptions.InvalidHttpRefererError
        pass

    def failed_to_authenticate_handler(self, error_obj):
        self.response['Content-Type'] = 'text/plain'
        self.response.content = error_obj.message

    def default_error_handler(self, error_obj):
        self.response['Content-Type'] = 'text/plain'
        self.response.content = error_obj.message

    def lack_of_required_key(self, error_obj):
        self.default_error_handler(error_obj)

    def key_validation_failed(self, error_obj):
        self.default_error_handler(error_obj)

    def not_allowed_method(self, *args, **kwargs):
        raise MethodNotAllowedError(self.request)



from django import VERSION as DJANGO_VERSION
from django.http import QueryDict

class PutRequestMiddleware(object):
    def process_request(self, request):
        if request.method.upper() == 'PUT':
            if not hasattr(request, 'PUT'):
                if DJANGO_VERSION >= (1,4):
                    raw_body = request.body
                else:
                    raw_body = request.raw_post_data
                setattr(request, request.method.upper(), QueryDict(raw_body))
        elif request.method.upper() not in ('POST', 'GET'):
            setattr(request, request.method.upper(), request.GET)



"""
Copyright 2012 Dian-Je Tsai and Wantoto Inc

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import imp
import os
import types
import urllib
import urlparse
from django.core.exceptions import ValidationError
from django.core.validators import validate_email
from six import callable


def valid_email(email):
    try:
        validate_email(email)
        return True
    except ValidationError:
        return False


def append_query_string_to_URL(url, data):
    url_parts = list(urlparse.urlparse(url))
    query = dict(urlparse.parse_qsl(url_parts[4]))

    if isinstance(data, str):
        data = dict(urlparse.parse_qsl(data))
    query.update(data)

    url_parts[4] = urllib.urlencode(query)

    return urlparse.urlunparse(url_parts)


def sql_string_from_query_set(qs, *args):
    sql = str(qs.query)
    for target, newString in args:
        sql = sql.replace(target, newString)
    return sql


def import_from_string(path):
    path_components = path.split('.')
    class_name = path_components[-1]
    package = '.'.join(path_components[:-1])
    return getattr(__import__(package, fromlist=[class_name]), class_name)


def dict_from_module(module_object, key_filter=None):
    if key_filter is None:
        key_filter = lambda x: True

    result = {}
    for attribute in dir(module_object):
        if key_filter(attribute):
            result[attribute] = getattr(module_object, attribute)
    return result


class SettingsDict(object):
    def __init__(self, default_settings, user_settings, **patches):
        if isinstance(default_settings, (str, unicode)):
            default_settings_name = os.path.split(os.path.splitext(default_settings)[0])[-1]
            default_settings = imp.load_source(default_settings_name, default_settings)
        if isinstance(default_settings, types.ModuleType):
            default_settings = dict_from_module(default_settings, lambda key: key.isupper())

        if isinstance(user_settings, (str, unicode)):
            user_settings_name = os.path.split(os.path.splitext(user_settings)[0])[-1]
            user_settings = imp.load_source(user_settings_name, user_settings)
        if isinstance(user_settings, types.ModuleType):
            user_settings = dict_from_module(user_settings, lambda key: key.isupper())

        settings = dict(default_settings)
        settings.update(user_settings)
        settings.update(patches)

        self._settings = settings

    def get(self, *args):
        if len(args) == 1:
            return self._get(args[0])
        elif len(args) == 2:
            return self._get_with_default(args[0], args[1])
        else:
            raise TypeError('get() takes exactly 1 or 2 arguments (%s given)' % len(args))

    def _get(self, key):
        if key in self._settings:
            target = self._settings[key]
            if callable(target):
                return target()
            else:
                return target

        else:
            raise KeyError(key)

    def _get_with_default(self, key, default):
        try:
            return self._get(key)
        except KeyError:
            return default

    def __getattr__(self, item):
        return self.get(item)



from django_eggplant import settings
from eggplant.constants.settings import AJAX_SIGNATURE_KEY_CLASS
from eggplant.utils import import_from_string

def ajax_signature_key(request):
    key_class_path = getattr(settings, AJAX_SIGNATURE_KEY_CLASS, None)

    if key_class_path is not None:
        AjaxSignatureKey = import_from_string(key_class_path)
    else:
        from eggplant.auth.signature import AjaxSignatureKey

    return {'AJAX_REQUEST_SIGNATURE_KEY': AjaxSignatureKey().generator(request)}


"""
Copyright 2012 Dian-Je Tsai, Cramdroid, and Wantoto Inc

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import yajl as json
from django.conf import settings
from django.contrib.auth import authenticate, login
from django.views.decorators.csrf import csrf_exempt
from eggplant.constants.settings import DEPLOYED, SIGNATURE_KEY_CLASS, AJAX_SIGNATURE_KEY_CLASS, DEFAULT_AUTH_METHOD
from eggplant.datatype import AttributedDict
from eggplant.utils import import_from_string
from eggplant.views import View, exceptions

class APIView(View):
    AUTH_METHODS = AttributedDict({'NoAuth': 0, 'Session': 1, 'Token': 2})

    @csrf_exempt
    def handle(self, request, *args, **kwargs):
        return super(APIView, self).handle(request, *args, **kwargs)

    def only_accept_https(self, *args, **kwargs):
        return getattr(settings, DEPLOYED, True)

    def should_check_signature_key(self, *args, **kwargs):
        return True

    def user_authentication_method(self, *args, **kwargs):
        default_auth_method = getattr(settings, DEFAULT_AUTH_METHOD, None)
        if default_auth_method is not None:
            return default_auth_method

        if self.init_kwargs.has_key('auth_method'):
            return self.init_kwargs['auth_method']

        if self.request.is_ajax():
            return APIView.AUTH_METHODS.Session
        else:
            return APIView.AUTH_METHODS.Token

    def signature_key(self, *args, **kwargs):
        return self.request.META.get('HTTP_X_REQUEST_SIGNATURE', '')

    def signature_key_class(self, *args, **kwargs):
        key_class_path = getattr(settings, SIGNATURE_KEY_CLASS, None)
        if self.request.is_ajax():
            key_class_path = getattr(settings, AJAX_SIGNATURE_KEY_CLASS, key_class_path)

        if key_class_path is not None:
            SignatureKey = import_from_string(key_class_path)
        else:
            from eggplant.auth.signature import SignatureKeyBase as SignatureKey
        return SignatureKey

    def user_token(self, *args, **kwargs):
        if self.request.is_ajax():
            return getattr(self.request, self.request.method.upper()).get('token', None)
        else:
            return self.request.META.get('HTTP_X_USER_TOKEN', None)

    def required_permissions(self, *args, **kwargs):
        return None

    def authenticate(self, *args, **kwargs):
        # Check app identification
        if self.should_check_signature_key(*args, **kwargs):
            SignatureKey = self.signature_key_class(*args, **kwargs)
            self.signature_key_handler = SignatureKey()

            key = self.signature_key(*args, **kwargs)
            key_is_valid = self.signature_key_handler.validator(self.request.path, key)
            if not key_is_valid:
                raise exceptions.SignatureKeyInvalidError()

            # Check permission
            required_permissions = self.required_permissions(*args, **kwargs)
            if required_permissions is not None:
                if not self.signature_key_handler.have_permission(key, required_permissions):
                    raise exceptions.AppHaveNoPermissionError()

        # Check user identification
        user_auth_method = self.user_authentication_method(*args, **kwargs)
        should_check_user_permission = True
        if user_auth_method==self.AUTH_METHODS.Session:
            if not self.request.user.is_authenticated():
                raise exceptions.UserIdentityUnknownError()
        elif user_auth_method==self.AUTH_METHODS.Token:
            token = self.user_token(*args, **kwargs)
            user = authenticate(token=token)
            if user is None:
                raise exceptions.UserIdentityUnknownError()
            else:
                login(self.request, user)
        else:
            should_check_user_permission = False

        # Check user's permission
        if should_check_user_permission:
            request_method = self.request.method.lower()
            if (request_method=='get'    and not self.user_can_read(*args, **kwargs)) or\
               (request_method=='post'   and not self.user_can_create(*args, **kwargs)) or\
               (request_method=='put'    and not self.user_can_update(*args, **kwargs)) or\
               (request_method=='delete' and not self.user_can_delete(*args, **kwargs)):
                raise exceptions.UserHaveNoPermissionError()

    def user_can_read(self, *args, **kwargs):   return True
    def user_can_create(self, *args, **kwargs): return False
    def user_can_update(self, *args, **kwargs): return False
    def user_can_delete(self, *args, **kwargs): return False

    def default_error_handler(self, error_obj):
        self.response['Content-Type'] = 'application/json'
        result = {
            'http_status_reason': error_obj.reason,
            'http_status_code': error_obj.status_code,
            'error': error_obj.message,
        }
        meta = {}
        if error_obj.user_info and len(error_obj.user_info.items()): meta['user_info'] = error_obj.user_info
        if error_obj.init_kwargs and len(error_obj.init_kwargs.items()): meta['keyword'] = error_obj.init_kwargs

        if len(meta.items()):
            try: result['meta'] = json.loads(json.dumps(meta))
            except TypeError: pass
        self.response.content = json.dumps(result)

    def post_handle(self, *args, **kwargs):
        self.response['Content-Type'] = 'application/json'


class DevAPIView(APIView):
    def only_accept_https(self, *args, **kwargs):
        return False

    def should_check_signature_key(self, *args, **kwargs):
        return False

    def user_authentication_method(self, *args, **kwargs):
        return APIView.AUTH_METHODS.NoAuth



"""
Copyright (c) 2012, Derek Schaefer.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1) Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.
    2) Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.
    3) Neither the name of django-json-field nor the names of its contributors
       may be used to endorse or promote products derived from this software
       without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

https://github.com/derek-schaefer/django-json-field/

"""

from eggplant.forms.fields.json_field import JSONFormField

from django.db import models
from django.utils import simplejson as json
from django.core import exceptions
from django.core.serializers.json import DjangoJSONEncoder
from django.utils.translation import ugettext as _
from django.core.exceptions import ImproperlyConfigured

import re
from decimal import Decimal
from datetime import datetime
try:
    from dateutil import parser as date_parser
except ImportError:
    raise ImproperlyConfigured('The "dateutil" library is required and was not found.')

try:
    JSON_DECODE_ERROR = json.JSONDecodeError # simplejson
except AttributeError:
    JSON_DECODE_ERROR = ValueError # other

TIME_RE = re.compile(r'^\d{2}:\d{2}:\d{2}')
DATE_RE = re.compile(r'^\d{4}-\d{2}-\d{2}(?!T)')
DATETIME_RE = re.compile(r'^\d{4}-\d{2}-\d{2}T')

class JSONDecoder(json.JSONDecoder):
    """ Recursive JSON to Python deserialization. """

    _recursable_types = [str, unicode, list, dict]

    def _is_recursive(self, obj):
        return type(obj) in JSONDecoder._recursable_types

    def decode(self, obj, *args, **kwargs):
        if not kwargs.get('recurse', False):
            obj = super(JSONDecoder, self).decode(obj, *args, **kwargs)
        if isinstance(obj, list):
            for i in xrange(len(obj)):
                item = obj[i]
                if self._is_recursive(item):
                    obj[i] = self.decode(item, recurse=True)
        elif isinstance(obj, dict):
            for key, value in obj.items():
                if self._is_recursive(value):
                    obj[key] = self.decode(value, recurse=True)
        elif isinstance(obj, basestring):
            if TIME_RE.match(obj):
                try:
                    return date_parser.parse(obj).time()
                except ValueError:
                    pass
            if DATE_RE.match(obj):
                try:
                    return date_parser.parse(obj).date()
                except ValueError:
                    pass
            if DATETIME_RE.match(obj):
                try:
                    return date_parser.parse(obj)
                except ValueError:
                    pass
        return obj

class JSONField(models.TextField):
    """ Stores and loads valid JSON objects. """

    __metaclass__ = models.SubfieldBase

    description = 'JSON object'

    def __init__(self, *args, **kwargs):
        self.default_error_messages = {
            'invalid': _(u'Enter a valid JSON object')
        }
        self._db_type = kwargs.pop('db_type', None)
        self.simple_formfield = kwargs.pop('simple_formfield', False)

        encoder = kwargs.pop('encoder', DjangoJSONEncoder)
        decoder = kwargs.pop('decoder', JSONDecoder)
        encoder_kwargs = kwargs.pop('encoder_kwargs', {})
        decoder_kwargs = kwargs.pop('decoder_kwargs', {})
        if not encoder_kwargs and encoder:
            encoder_kwargs.update({'cls':encoder})
        if not decoder_kwargs and decoder:
            decoder_kwargs.update({'cls':decoder, 'parse_float':Decimal})
        self.encoder_kwargs = encoder_kwargs
        self.decoder_kwargs = decoder_kwargs

        kwargs['default'] = kwargs.get('default', 'null')
        kwargs['help_text'] = kwargs.get('help_text', self.default_error_messages['invalid'])

        super(JSONField, self).__init__(*args, **kwargs)

    def db_type(self, *args, **kwargs):
        if self._db_type:
            return self._db_type
        return super(JSONField, self).db_type(*args, **kwargs)

    def to_python(self, value):
        if value is None: # allow blank objects
            return None
        if isinstance(value, basestring):
            try:
                value = json.loads(value, **self.decoder_kwargs)
            except JSON_DECODE_ERROR:
                pass
        return value

    def get_db_prep_value(self, value, *args, **kwargs):
        if self.null and value is None and not kwargs.get('force'):
            return None
        return json.dumps(value, **self.encoder_kwargs)

    def value_to_string(self, obj):
        return self.get_db_prep_value(self._get_val_from_obj(obj))

    def value_from_object(self, obj):
        return json.dumps(super(JSONField, self).value_from_object(obj), **self.encoder_kwargs)

    def formfield(self, **kwargs):
        defaults = {
            'form_class': kwargs.get('form_class', JSONFormField),
            'simple': self.simple_formfield,
            'encoder_kwargs': self.encoder_kwargs,
            'decoder_kwargs': self.decoder_kwargs,
        }
        defaults.update(kwargs)
        return super(JSONField, self).formfield(**defaults)

    def contribute_to_class(self, cls, name):
        super(JSONField, self).contribute_to_class(cls, name)

        def get_json(model_instance):
            return self.get_db_prep_value(getattr(model_instance, self.attname, None), force=True)
        setattr(cls, 'get_%s_json' % self.name, get_json)

        def set_json(model_instance, value):
            return setattr(model_instance, self.attname, self.to_python(value))
        setattr(cls, 'set_%s_json' % self.name, set_json)

try:
    # add support for South migrations
    from south.modelsinspector import add_introspection_rules
    #add_introspection_rules([], ['^json_field\.fields\.JSONField'])
    add_introspection_rules([], ['^eggplant\.models\.fields\.json_field\.JSONField'])
except ImportError:
    pass



"""
Copyright 2012 Dian-Je Tsai and Wantoto Inc

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from functools import wraps
import types
from django import VERSION as DJANGO_VERSION
from django.core.exceptions import ValidationError
from django.http import QueryDict, HttpRequest
import six
from eggplant.views.exceptions import LackOfRequiredKeyError, KeyValidationError
from eggplant.views.functions import lack_of_required_key_default_handler, key_validation_failed_default_handler
from eggplant.views import View

def _get_request_data(request):
    if request.method.lower() == 'put':
        if hasattr(request, request.method.upper()):
            request_data = request.PUT
        else:
            if DJANGO_VERSION[1] >= 4:
                raw_request_data = request.body
            else:
                raw_request_data = request.raw_post_data
            request_data = QueryDict(raw_request_data)
            setattr(request, request.method.upper(), request_data)
    elif request.method.lower() == 'post':
        request_data = request.POST
    else:
        request_data = request.GET

    return request_data


def _get_wrapped_type(args):
    if isinstance(args[0], HttpRequest):
        # Function base
        view_object = None
        request = args[0]
        is_class_based = False
    elif isinstance(args[0], View):
        # Class base
        view_object = args[0]
        request = args[0].request
        is_class_based = True
    else:
        view_object = None
        request = None
        is_class_based = None

    return view_object, request, is_class_based


def _normalize_key_validator_pairs(raw_pairs):
    key_validator_pairs = list(raw_pairs)
    if len(raw_pairs)==2:
        if isinstance(raw_pairs[0], six.string_types):
            key_validator_pairs[0] = [raw_pairs[0]]
        if isinstance(raw_pairs[1], (types.FunctionType, staticmethod)):
            key_validator_pairs[1] = [raw_pairs[1]]

        is_key_list = True
        for item in key_validator_pairs[0]:
            if not isinstance(item, six.string_types):
                is_key_list = False
                break
        is_validator_list = True
        for item in key_validator_pairs[1]:
            if not isinstance(item, (types.FunctionType, staticmethod)):
                is_validator_list = False
                break
        if is_key_list and is_validator_list:
            key_validator_pairs = (key_validator_pairs,)
    return key_validator_pairs


def required_keys(*keys, **kwargs):
    error_handler = kwargs.get('error_handler', None)
    if error_handler is None:
        error_handler = lack_of_required_key_default_handler

    def wrapper(fn):
        @wraps(fn)
        def wrapped(*args, **kwargs):
            view_object, request, is_class_based = _get_wrapped_type(args)
            if request is None:
                return fn(*args, **kwargs)

            request_data = _get_request_data(request)

            for key in keys:
                if key not in request_data:
                    if is_class_based:
                        view_object.response.content = ''
                        e = LackOfRequiredKeyError(key)
                        e.error_handler = view_object.lack_of_required_key
                        raise e
                    else:
                        return error_handler(request, key, *args, **kwargs)

            return fn(*args, **kwargs)

        return wrapped

    return wrapper


def validate_keys(*raw_key_validator_tuple, **kwargs):
    error_handler = kwargs.get('error_handler', None)
    if error_handler is None:
        error_handler = key_validation_failed_default_handler

    def wrapper(fn):
        @wraps(fn)
        def wrapped(*args, **kwargs):
            view_object, request, is_class_based = _get_wrapped_type(args)
            if request is None:
                return fn(*args, **kwargs)

            request_data = _get_request_data(request)

            key_validator_pairs = _normalize_key_validator_pairs(raw_key_validator_tuple)

            for request_key, validator in key_validator_pairs:
                if not isinstance(request_key, (list, tuple)): keys = [request_key]
                else: keys = request_key

                if not isinstance(validator, (list, tuple)): validators = [validator]
                else: validators = validator

                for key in keys:
                    for validator in validators:
                        if key in request_data:
                            try:
                                if is_class_based:
                                    if isinstance(validator, staticmethod):
                                        validator_name = validator.__func__.__name__
                                        arguments = (request_data[key],)
                                    else:
                                        # ==> isinstance(validator, types.FunctionType)
                                        validator_name = validator.__name__
                                        arguments = (view_object, request_data[key])

                                    if hasattr(view_object, validator_name):
                                        getattr(view_object.__class__, validator_name)(*arguments)
                                    else:
                                        validator(request_data[key])
                                else:
                                    validator(request_data[key])
                            except ValidationError, e:
                                if is_class_based:
                                    view_object.response.content = ''
                                    e = KeyValidationError(key, e.messages)
                                    e.error_handler = view_object.key_validation_failed
                                    raise e
                                else:
                                    return error_handler(request, key, *args, **kwargs)

            return fn(*args, **kwargs)

        return wrapped

    return wrapper



"""
Copyright 2012 Dian-Je Tsai and Wantoto Inc

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import six
if six.PY3:
    import http.client as httplib
else:
    import httplib

class ViewError(Exception):
    def __init__(self, status_code, message=None, **kwargs):
        self.status_code = status_code
        self.reason = httplib.responses[status_code]
        self.message = message or 'Something goes wrong ...'
        self.error_handler = None
        self.user_info = None
        self.init_kwargs = kwargs

#-----------------------------------------------------------------------------------------------------------------------
# 400 Bad Request
# Request syntax error (Lack of required keys, invalid value of arguments)

class BadRequestError(ViewError):
    def __init__(self, message=None, **kwargs):
        super(BadRequestError, self).__init__(httplib.BAD_REQUEST, message or 'Bad Request', **kwargs)

class LackOfRequiredKeyError(BadRequestError):
    def __init__(self, key, **kwargs):
        super(LackOfRequiredKeyError, self).__init__('Key "%s" is required.'%key, **kwargs)
        self.user_info = {'key': key}

class KeyValidationError(BadRequestError):
    def __init__(self, key, messages, **kwargs):
        message = ('Key "%s" is invalid.' % key + ' ' + ' '.join(messages)).strip()
        super(KeyValidationError, self).__init__(message, **kwargs)
        self.user_info = {
            'key': key,
            'raw_messages': messages
        }

#-----------------------------------------------------------------------------------------------------------------------
# 401 Unauthorized
# Failed to authenticate user. Credential error or lack of required credential

class UnauthorizedError(ViewError):
    def __init__(self, message=None, **kwargs):
        super(UnauthorizedError, self).__init__(httplib.UNAUTHORIZED,
            message or 'You have to login before access this resources.', **kwargs)

class SignatureKeyInvalidError(UnauthorizedError):
    def __init__(self, **kwargs):
        super(SignatureKeyInvalidError, self).__init__('The signature key is invalid.', **kwargs)

class UserIdentityUnknownError(UnauthorizedError):
    def __init__(self, **kwargs):
        super(UserIdentityUnknownError, self).__init__('Your identity is unknown.', **kwargs)

#-----------------------------------------------------------------------------------------------------------------------
# 403 Forbidden
# Authenticated but don't have enough permission

class ForbiddenError(ViewError):
    def __init__(self, message=None, **kwargs):
        super(ForbiddenError, self).__init__(httplib.FORBIDDEN,
            message or 'You have no permission to access this resources.', **kwargs)

class InvalidHttpRefererError(ForbiddenError):
    def __init__(self, **kwargs):
        super(InvalidHttpRefererError, self).__init__('This http refer of this request is not accepted.', **kwargs)

class AppHaveNoPermissionError(ForbiddenError):
    def __init__(self, **kwargs):
        super(AppHaveNoPermissionError, self).__init__(
            'This application doesn\'t have permission to access this.', **kwargs)

class UserHaveNoPermissionError(ForbiddenError):
    def __init__(self, **kwargs):
        super(UserHaveNoPermissionError, self).__init__('You have no permission to access this.', **kwargs)

#-----------------------------------------------------------------------------------------------------------------------
# 404 Not Found

class NotFoundError(ViewError):
    def __init__(self, message=None, **kwargs):
        super(NotFoundError, self).__init__(httplib.NOT_FOUND, message or 'Item not found', **kwargs)

class ModelDoesNotExistError(NotFoundError):
    def __init__(self, model_name, **kwargs):
        super(ModelDoesNotExistError, self).__init__('No such %s item.'%model_name, **kwargs)

#-----------------------------------------------------------------------------------------------------------------------
# 405 Method Not Allowed

class MethodNotAllowedError(ViewError):
    def __init__(self, http_request, message=None, **kwargs):
        super(MethodNotAllowedError, self).__init__(httplib.METHOD_NOT_ALLOWED,
            message or 'HTTP %s Method is not allowed.'%http_request.method.upper(), **kwargs)

#-----------------------------------------------------------------------------------------------------------------------
# 409 Conflict

class ConflictError(ViewError):
    def __init__(self, message=None, **kwargs):
        super(ConflictError, self).__init__(httplib.CONFLICT, message or 'Conflict', **kwargs)

class DuplicatedInsertionError(ConflictError):
    def __init__(self, message=None, **kwargs):
        super(DuplicatedInsertionError, self).__init__(
            message or 'Requested insertion contains duplicated value for unique keys.', **kwargs)



"""
Copyright 2012 Dian-Je Tsai and Wantoto Inc

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from fabric.api import abort, env, sudo, execute, puts
from fabric.colors import red
from fabric.decorators import task, roles
from netaddr import IPNetwork
from netaddr.core import AddrFormatError


@task
@roles('web_servers')
def apache(action, load_deploy_settings=False, project_name=None, deploy_settings_path=None):
    """Perform action on apache server
    :param action: action could be ('start', 'stop', 'restart', 'reload', 'status', 'fullstatus')
    """
    if load_deploy_settings:
        execute('load_deploy_setting', project_name=project_name, deploy_settings_path=deploy_settings_path)
        deploy_settings = env.deploy_settings
        execute(web_server, *deploy_settings.WEB_SERVERS)

    if action not in ('start', 'stop', 'restart', 'reload', 'status', 'fullstatus'):
        abort(red('"%s"' % action, bold=True) + red(' is an invalid action for apache httpd!'))
    sudo('/etc/init.d/httpd %s' % action)


@task
def web_server(*servers):
    """Set targeted web server
    :param servers: list of severs (IP, CIDR, or domain name)
    """
    hosts = []
    for server in servers:
        try:
            hosts += map(lambda x: str(x), list(IPNetwork(server)))
        except (ValueError, AddrFormatError):
            hosts += [server]

    user_hosts = []
    for host in hosts:
        puts('Add "%s" to web server list' % host)
        user_hosts.append('eggplant@%s' % host)

    env.roledefs.update({
        'web_servers': env.roledefs['web_servers'] + user_hosts
    })


@task
def setup(target_server):
    print target_server



"""
Copyright 2012 Dian-Je Tsai and Wantoto Inc

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import re
from django.core.exceptions import ValidationError
from eggplant.utils.string import str2bool

def is_hex(value):
    if re.compile(r'^[a-fA-F0-9]*$').match(value) is None:
        raise ValidationError('%s is not a valid hex string'%value)

def is_integer(raw_value):
    try:
        return int(raw_value)
    except ValueError:
        raise ValidationError('\'%s\' is not an integer.'%raw_value)

def is_positive_integer(raw_value):
    int_value = is_integer(raw_value)
    if int_value <= 0:
        raise ValidationError('\'%s\' is not a positive integer.'%raw_value)
    return int_value

def is_natural_integer(raw_value):
    int_value = is_integer(raw_value)
    if int_value < 0:
        raise ValidationError('\'%s\' is not a natural integer.'%raw_value)
    return int_value

def is_float(raw_value):
    try:
        return float(raw_value)
    except ValueError:
        raise ValidationError('\'%s\' is not a float number.'%raw_value)

def is_positive_float(raw_value):
    float_value = is_float(raw_value)
    if float_value <= 0:
        raise ValidationError('\'%s\' is not a positive float number.'%raw_value)
    return float_value

def is_natural_float(raw_value):
    float_value = is_float(raw_value)
    if float_value < 0:
        raise ValidationError('\'%s\' is not a natural float.'%raw_value)
    return float_value

def is_boolean(raw_value):
    try:
        return str2bool(raw_value)
    except ValueError:
        raise ValidationError('\'%s\' is not a valid boolean value.'%raw_value)




"""
http://djangosnippets.org/snippets/2282/

Allows url patterns to include a boolean indicating whether a view requires
TLS(SSL).  The accompanying middleware handles the redirects needed to make
sure that it upholds this requirement.

WARNING: this monkey-patches some Django internals and is difficult to test
since Django's TestClient does not support TLS.  If you use this make sure you
test it thouroughly.

Add this to your Django settings::

    USE_TLS = True   # The default for this setting is False.

URL pattern usage::

    url(r'^login$', 'myproject.login.index',
        {'require_tls': True}, name='login-index'),

Use ``require_tls`` True to force the middleware to perform redirects needed to
make sure your are serving this view using https.

Use ``require_tls`` False to force the middleware to redirect to http.  Be
careful with this setting, this may not behave as you expect.  If you don't
care if the view is served via https or http then do not include
``require_tls`` in the pattern.

If you wish to have every view in the site served with TLS then specify the
following Django setting::

    ALWAYS_USE_TLS = True   # Django setting, use TLS for every view

"""
from django.conf import settings
from django.core import urlresolvers
from django.http import HttpResponsePermanentRedirect
from django.http import get_host
from django.utils.encoding import smart_str

ALWAYS_USE_TLS = getattr(settings, 'ALWAYS_USE_TLS', False)
USE_TLS = getattr(settings, 'HTTPS_SUPPORT', ALWAYS_USE_TLS)


def resolve_pattern(path, urlconf=None):
    """
    Works the same as Django's url resolver but instead of return a view
    function that handles a specific path it gives you the pattern that matches
    it
    """
    return urlresolvers.get_resolver(urlconf).resolve_pattern(path)


def resolve_pattern_method(self, path):
    """
    Monkey-patchable to the RegexURLResolver class

    The difference is that this method will return the pattern instead of what
    Django's RegexURLResolver.resolve returns
    """
    tried = []
    match = self.regex.search(path)
    if match:
        new_path = path[match.end():]
        for pattern in self.url_patterns:
            try:
                if hasattr(pattern, 'resolve_pattern'):
                    # Use our resolve_pattern method
                    sub_match = pattern.resolve_pattern(new_path)
                    if isinstance(sub_match, urlresolvers.RegexURLPattern):
                        return sub_match
                else:
                    sub_match = pattern.resolve(new_path)
            except urlresolvers.Resolver404, e:
                sub_tried = e.args[0].get('tried')
                if sub_tried is not None:
                    tried.extend([(pattern.regex.pattern + '   ' + t) for t in sub_tried])
                else:
                    tried.append(pattern.regex.pattern)
            else:
                if sub_match:
                    if hasattr(pattern, 'callback'):
                        # Found our pattern
                        return pattern
                    sub_match_dict = dict([(smart_str(k), v) for k, v in match.groupdict().items()])
                    sub_match_dict.update(self.default_kwargs)
                    for k, v in sub_match[2].iteritems():
                        sub_match_dict[smart_str(k)] = v
                    return sub_match[0], sub_match[1], sub_match_dict
                tried.append(pattern.regex.pattern)
        raise urlresolvers.Resolver404, {'tried': tried, 'path': new_path}
    raise urlresolvers.Resolver404, {'path' : path}

# Monkey patch the additional methods
urlresolvers.RegexURLResolver.resolve_pattern = resolve_pattern_method


def reverse_decorator(func):
    def inner(*args, **kwargs):
        if 'request' in kwargs:
            request = kwargs['request']
            del kwargs['request']
        else:
            request = None

        # Call the real reverse function
        abs_path = func(*args, **kwargs)

        if not request or not USE_TLS:
            # Short-circuit, no need to continue becuase there is no request we
            # can get smart with or TLS is turned off
            return abs_path

        # Let's figure out if we need this URL to be secure or not
        pattern = resolve_pattern(abs_path)
        if 'require_tls' in pattern.default_args:
            secure = pattern.default_args['require_tls']
        else:
            secure = None

        if not secure is None and request and \
           not secure == is_secure(request):
            # Secure indicates either http or https and the request is not
            # currently being served that way
            protocol = 'https' if secure else 'http'
            abs_path = '%s://%s%s' % (protocol, get_host(request), abs_path,)
        return abs_path
    return inner
urlresolvers.reverse = reverse_decorator(urlresolvers.reverse)


def is_secure(request):
    """
    Determines if the given request is over HTTPS or not
    """
    if request.is_secure():
        return True

    return False


class TlsRedirectMiddleware(object):
    def process_view(self, request, view_func, view_args, view_kwargs):
        if 'require_tls' in view_kwargs:
            secure = view_kwargs['require_tls']
            del view_kwargs['require_tls']
        else:
            secure = None

        if not any([USE_TLS, ALWAYS_USE_TLS]):
            return None

        if ALWAYS_USE_TLS:
            # This site is setup to always use TLS so we set the secure bit to
            # always be true
            secure = True

        # If secure is None, the require_tls was not present in the view
        # keyword args at all.  In this case, we don't want to redirect.
        # Leaving this keyword off probably indicates that the particular url
        # doesn't require http or https, we go with whatever we got
        if not secure is None and \
           not secure == is_secure(request):
            return self._redirect(request, secure)

    def _redirect(self, request, secure):
        #protocol = secure and 'https' or 'http'
        protocol = ''
        if secure: protocol = 'https'
        else: protocol = 'http'

        newurl = '%s://%s%s' % (
            protocol, get_host(request), request.get_full_path(),)
        # replace for dev server
        if settings.DEBUG and secure:
            newurl = newurl.replace(':8000', ':8443')
        if settings.DEBUG and request.method == 'POST':
            raise RuntimeError('Django can\'t perform a TLS redirect while '
                'maintaining POST data.  Please structure your views so that '
                'redirects only occur during GETs.')

        return HttpResponsePermanentRedirect(newurl)



import datetime
import sys
from pytz import timezone
from django.conf import settings

current_timezone = timezone(settings.TIME_ZONE)

def now_with_timezone():
    return datetime.datetime.now().replace(tzinfo=current_timezone)

def total_seconds(td):
    if sys.version_info < (2 ,7):
        # timedelta has no total_seconds
        return (td.microseconds + (td.seconds + td.days * 24 * 3600) * 10**6) / float(10**6)
    else:
        return td.total_seconds()



from urllib import urlencode
import urlparse

def replace_querystring(original_url, **replacement):
    url_components = urlparse.urlparse(original_url)
    query_strings = urlparse.parse_qs(url_components.query)

    for key, value in replacement.items():
        query_strings[key] = value

    new_url_components = url_components._replace(query=urlencode(query_strings, True))
    new_url = urlparse.urlunparse(new_url_components)

    return new_url




from django.conf import settings
from django.contrib.auth.backends import ModelBackend
from eggplant.constants.settings import USER_TOKEN_CLASS
from eggplant.utils import import_from_string

class TokenBackend(ModelBackend):
    def authenticate(self, token=None):
        token_class_path = getattr(settings, USER_TOKEN_CLASS, 'eggplant.auth.token.UserToken')
        UserToken = import_from_string(token_class_path)

        # Check User
        user_token_handler = UserToken()
        if user_token_handler.validator(token):
            return user_token_handler.user_from_token(token)
        else:
            return None


"""
Copyright (c) 2012, Derek Schaefer.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1) Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.
    2) Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.
    3) Neither the name of django-json-field nor the names of its contributors
       may be used to endorse or promote products derived from this software
       without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

https://github.com/derek-schaefer/django-json-field/

"""

from django.forms import fields, util
from django.utils import simplejson as json

import datetime

class JSONFormField(fields.Field):

    def __init__(self, *args, **kwargs):
        self.simple = kwargs.pop('simple', False)
        self.encoder_kwargs = kwargs.pop('encoder_kwargs', {})
        self.decoder_kwargs = kwargs.pop('decoder_kwargs', {})
        super(JSONFormField, self).__init__(*args, **kwargs)

    def clean(self, value):
        # Have to jump through a few hoops to make this reliable
        value = super(JSONFormField, self).clean(value)

        # allow an empty value on an optional field
        if value is None:
            return value

        ## Got to get rid of newlines for validation to work
        # Data newlines are escaped so this is safe
        value = value.replace('\r', '').replace('\n', '')

        json_globals = { # safety first!
                         '__builtins__': None,
        }
        if not self.simple: # optional restriction
            json_globals.update({'datetime':datetime})
        json_locals = { # value compatibility
                        'null': None,
                        'true': True,
                        'false': False,
        }
        try:
            value = json.dumps(eval(value, json_globals, json_locals), **self.encoder_kwargs)
        except Exception, e: # eval can throw many different errors
            raise util.ValidationError('%s (Caught "%s")' % (self.help_text, e))

        try:
            json.loads(value, **self.decoder_kwargs)
        except ValueError, e:
            raise util.ValidationError('%s (Caught "%s")' % (self.help_text, e))

        return value



# https://github.com/rhec/django-html5
from django.forms.fields import IntegerField as DjangoIntegerField
from eggplant.forms.widgets.html5 import NumberInput

class IntegerField(DjangoIntegerField):
    widget = NumberInput

    def widget_attrs(self, widget):
        """
        Given a Widget instance (*not* a Widget class), returns a dictionary of
        any HTML attributes that should be added to the Widget, based on this
        Field.
        """
        attrs = {}
        if self.min_value is not None:
            attrs['min'] = self.min_value
        if self.max_value is not None:
            attrs['max'] = self.max_value
        return attrs


"""
Copyright 2012 Dian-Je Tsai and Wantoto Inc

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from django.db import models
from eggplant import APP_LABEL
from eggplant.utils.json import json_model_dict

class DictModel(models.Model):

    """
    An abstract model to make Models have dictionary-like interface
    """

    def __getitem__(self, key):
        if key in self.fields:
            return getattr(self, key)
        else:
            raise KeyError, '%s is not a field.' % key

    def __setitem__(self, key, value):
        if key in self.fields:
            return setattr(self, key, value)
        else:
            raise KeyError, '%s is not a field.' % key

    def __contains__(self, item):
        return item in self.keys()

    def has_key(self, k):
        return k in self.keys()

    def items(self):
        return [(field.name, self[field.name]) for field in self.__class__._meta.fields]

    def keys(self):
        return self.fields

    def values(self):
        return [self[field.name] for field in self.__class__._meta.fields]

    @property
    def fields(self):
        """Get a list containing all fields in this model
        :return: a list containing the name of fields in this model
        """
        if not hasattr(self, '_fields') or self._fields is None:
            self._fields = [field.name for field in self.__class__._meta.fields]
        return self._fields

    def dict(self, json_serializable=False, expand_choices=False, **kwargs):
        """Convert a model into dict
        :return: a dictionary containing fields name and values in this model
        """
        if json_serializable:
            json_serializable_fields = [field.name for field in self.__class__._meta.fields
                if not isinstance(field, (models.OneToOneField, models.ManyToManyField, models.ForeignKey))]
            result = dict([item for item in self.items() if item[0] in json_serializable_fields])
        else:
            result = dict(self.items())

        if expand_choices:
            for field, value in result.items():
                if hasattr(self, 'get_%s_display'%field.lower()):
                    result[field] = getattr(self, 'get_%s_display'%field.lower())()

        return result

    def json_dict(self, use_float_seconds=True, summary=False, **kwargs):
        """Convert a model into dict which is json serializable.

        DateTimeField and DateField will be converted into a unix-timestamp-like object.
        Standard unix-timestamp doesn't contain float seconds, but the one returned contains float seconds.

        TimeField will be converted into seconds from 0:00:00

        :return: a dictionary which is json serializable
        """
        return json_model_dict(self.dict(json_serializable=True, summary=summary, **kwargs),
            use_float_seconds=use_float_seconds)

    class Meta:
        abstract = True
        app_label = APP_LABEL



"""
Copyright 2012 Dian-Je Tsai, Cramdroid, and Wantoto Inc

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import re
from django.conf import settings
from django.core.cache import cache
from django.core.exceptions import ImproperlyConfigured, FieldError
from django.core.paginator import Paginator, EmptyPage
from django.db.models import Q
import yajl as json
from eggplant.utils.url import replace_querystring
from eggplant.validators.number import range_validator
from eggplant.validators.string import is_integer, is_positive_integer
from eggplant.views import exceptions
from eggplant.views.api import APIView
from eggplant.views.exceptions import ViewError
from eggplant.views.decorators import validate_keys
from eggplant.constants.settings import MAX_PAGINATION_SIZE, DEFAULT_PAGINATION_SIZE
from eggplant.constants import defaults as eggplant_defaults

FILTER_CACHE_KEY_PREFIX = 'com.wantoto.eggplant.views.api.model.filter'
ORDER_CACHE_KEY_PREFIX  = 'com.wantoto.eggplant.views.api.model.order'

SUPPORTED_FILTER_TYPES = [
    ('=',  '__exact'),
    ('i=', '__iexact'),
    ('>',  '__gt'),
    ('<',  '__lt'),
    ('>=', '__gte'),
    ('<=', '__lte'),
    ('%',  '__contains'),
    ('i%', '__icontains'),
    ('^',  '__startswith'),
    ('i^', '__istartswith'),
    ('$',  '__endswith'),
    ('i$', '__iendswith'),
    ('@', '__in'),
    ]
SUPPORTED_FILTER_TYPES_DICT = dict(SUPPORTED_FILTER_TYPES)

#-----------------------------------------------------------------------------------------------------------------------

class ModelViewBase(APIView):
    # Permission
    ACCEPT_GET = True
    ACCEPT_POST = False
    ACCEPT_PUT = False
    ACCEPT_DELETE = False

    def __init__(self, *args, **kwargs):
        super(ModelViewBase, self).__init__(*args, **kwargs)

        if not self.ACCEPT_GET and hasattr(self, 'get'):
            setattr(self, 'get', self.not_allowed_method)
        if not self.ACCEPT_POST and hasattr(self, 'post'):
            setattr(self, 'post', self.not_allowed_method)
        if not self.ACCEPT_PUT and hasattr(self, 'put'):
            setattr(self, 'put', self.not_allowed_method)
        if not self.ACCEPT_DELETE and hasattr(self, 'delete'):
            setattr(self, 'delete', self.not_allowed_method)

    def model_presenter(self, *args, **kwargs):
        return lambda x: getattr(x, 'json_dict')()

#-----------------------------------------------------------------------------------------------------------------------

class ModelCollectionView(ModelViewBase):
    MODEL_COLLECTION_CLASS = None
    ORDERABLE_FIELDS = []

    SKIP_DELIMITER_IN_FILTERS = False
    VALID_FILTERS_FOR_FIELDS = {}

    DEFAULT_PAGE_SIZE = getattr(settings, DEFAULT_PAGINATION_SIZE, eggplant_defaults.EGGPLANT_DEFAULT_PAGINATION_SIZE)

    #-------------------------------------------------------------------------------------------------------------------
    # Object Lifecycle

    def __init__(self, *args, **kwargs):
        super(ModelCollectionView, self).__init__(*args, **kwargs)

        if self.MODEL_COLLECTION_CLASS is None and kwargs.has_key('model_collection_class'):
            self.MODEL_COLLECTION_CLASS = kwargs['model_collection_class']

    #-------------------------------------------------------------------------------------------------------------------
    # Main methods

    @validate_keys(
        (('page_size', 'limit'),
          (is_positive_integer, range_validator(1, '>='),
           range_validator(getattr(settings, MAX_PAGINATION_SIZE, eggplant_defaults.EGGPLANT_MAX_PAGINATION_SIZE), '<=')
          )
         ),
        (('page_num', 'offset'), is_integer)
    )
    def get_collection(self, *args, **kwargs):
        # Cached one
        if hasattr(self, '_collection') and hasattr(self, '_queryset') and hasattr(self, '_meta_info'):
            return getattr(self, '_collection'), getattr(self, '_queryset'), getattr(self, '_meta_info')

        # Get query set
        query_set = self.queryset(*args, **kwargs)

        # Filter
        # "field:operation:value" with ;&| as connector
        # For example:
        #    name:=:broadcast;owner:=:sodas ==> name == broadcast AND owner == sodas
        #    name:=:broadcast&owner:=:sodas ==> name == broadcast AND owner == sodas
        #    name:=:broadcast|owner:=:sodas ==> name == broadcast OR owner == sodas
        #    name:=:broadcast;owner:=:sodas&year:=:2012 ==> name == broadcast AND (owner == sodas AND year == 2012)
        #    name:=:broadcast;owner:=:sodas|year:=:2012 ==> name == broadcast AND (owner == sodas OR year == 2012)
        #    name:=:broadcast;owner:=:sodas;year:=:2012 ==> name == broadcast AND owner == sodas AND year == 2012
        raw_filter = self.request.GET.get('filter', None)
        filters = None
        filter_cache_key = None
        if raw_filter is not None:
            # Get from cache
            filter_cache_key = '%s.%s'%(FILTER_CACHE_KEY_PREFIX, raw_filter)
            filters = cache.get(filter_cache_key)
        # Make new if cache is empty
        if raw_filter is not None and filters is None:
            filters = []
            # RE: LookAhead and LookBehind (?=...) (?!...), (?<=...) (?<!...)
            # Break ";" first
            and_group = re.split(r'(?<!\\);', raw_filter) if self.SKIP_DELIMITER_IN_FILTERS else raw_filter.split(';')
            for and_pair in and_group:
                # Get components separated by '&' or '|'
                # and_pair ==> 'field:op:value&field:op:value'
                # query_components ==> ['field:op:value', '&', 'field:op:value']
                # lookups ==> ['field:op:value', 'field:op:value']
                # connectors ==> ['&']
                if self.SKIP_DELIMITER_IN_FILTERS:
                    query_components = [item.replace('\\', '') for item in re.split(r'((?<!\\)&|(?<!\\)\|)', and_pair)]
                else:
                    query_components = re.split(r'(&|\|)', and_pair)
                lookups = query_components[::2] # a,b,c ==> a,c
                connectors = query_components[1::2] # a,b,c ==> b

                # Create each Q objects separated by '&' or '|'
                q_objects = []
                for lookup in lookups:
                    # Component
                    try:
                        field, filter_type, value = \
                            re.split(r'(?<!\\):', lookup) if self.SKIP_DELIMITER_IN_FILTERS else lookup.split(':')
                    except ValueError:
                        raise exceptions.BadRequestError('Error filter format: %s'%lookup)

                    # Field name and negate
                    negate = field.startswith('~')
                    if negate: field = field[1:]

                    # Check field name and filter type
                    if not self.VALID_FILTERS_FOR_FIELDS.has_key(field):
                        continue # Not a valid field
                    valid_filter_types = self.VALID_FILTERS_FOR_FIELDS[field]
                    if filter_type not in valid_filter_types:
                        continue # not a valid filter type

                    # Is value a list?
                    if filter_type=='@':
                        value = re.split(r'(?<!\\),', value) if self.SKIP_DELIMITER_IN_FILTERS else value.split(',')

                    # Create object
                    q_object = self.q_object_from_filter(field, filter_type, value, negate)
                    if q_object is not None: q_objects.append(q_object)

                # Connect each Q objects in a ',' via & or |
                final_q_object = None
                for index, q in enumerate(q_objects):
                    if index==0: final_q_object = q
                    else:
                        if connectors[index-1]=='&': final_q_object &= q
                        elif connectors[index-1]=='|': final_q_object |= q
                if final_q_object is not None: filters.append(final_q_object)
            # Write back to cache
            cache.set(filter_cache_key, filters)
        # Apply filter to query set
        try:
            if filters is not None and len(filters)!=0: query_set = query_set.filter(*filters)
        except FieldError, e:
            raise ImproperlyConfigured('Field is not valid: %s'%e.message)
        except ValueError:
            raise exceptions.BadRequestError('Error filter value')

        # Order
        raw_order = self.request.GET.get('order', None)
        order_cache_key = None
        orders = None
        if raw_order is not None:
            order_cache_key = '%s.%s'%(ORDER_CACHE_KEY_PREFIX, raw_order)
            orders = cache.get(order_cache_key)
        if raw_order is not None and orders is None:
            order_pairs = []
            raw_orders = raw_order.split(',')
            for raw_order in raw_orders:
                if raw_order.startswith('-'):
                    order = raw_order[1:]
                    direction = -1
                else:
                    order = raw_order
                    direction = 1
                if order in self.ORDERABLE_FIELDS:
                    order_pairs.append((direction, order))
            orders = self.order_list(query_set, order_pairs, *args, **kwargs)
            cache.set(order_cache_key, orders)
        try:
            if orders is not None and len(orders)!=0: query_set = query_set.order_by(*orders)
        except FieldError, e:
            raise ImproperlyConfigured('Field is not valid: %s'%e.message)

        meta_info = {}
        # Pagination
        default_page_size = self.DEFAULT_PAGE_SIZE
        if self.request.GET.has_key('page_size') or self.request.GET.has_key('page_num'):
            page_size = int(self.request.GET.get('page_size', default_page_size))
            current_page_num = int(self.request.GET.get('page_num', 1))

            paginator = Paginator(query_set, page_size)
            try:
                target_page = paginator.page(current_page_num)
            except EmptyPage:
                raise ViewError(400, 'Requested page number (%s) is out of range. (Range: %s:%s)'%(
                    current_page_num,
                    paginator.page_range[0],
                    paginator.page_range[-1],
                    ),
                    reason='WPNOutOfPage'
                )
            queryset = target_page.object_list

            meta_info['pagination'] = {
                'item_count': paginator.count,
                'page_count': paginator.num_pages,
                'page_size': page_size,
                'first_page': paginator.page_range[0],
                'last_page': paginator.page_range[-1],
                'current_page': current_page_num,
                'prev_page': target_page.previous_page_number() if target_page.has_previous() else None,
                'next_page': target_page.next_page_number() if target_page.has_next() else None,
            }
            if target_page.has_next():
                original_url = self.request.build_absolute_uri()
                new_url = replace_querystring(original_url, page_num=target_page.next_page_number())
                meta_info['pagination']['next_url'] = new_url
            if target_page.has_previous():
                original_url = self.request.build_absolute_uri()
                new_url = replace_querystring(original_url, page_num=target_page.previous_page_number())
                meta_info['pagination']['prev_url'] = new_url
        else:
            result_limit = int(self.request.GET.get('limit', default_page_size))
            result_offset = int(self.request.GET.get('offset', 0))

            item_count = query_set.count()
            queryset = query_set[result_offset:result_offset+result_limit]

            meta_info['pagination'] = {
                'item_count': item_count,
                'offset': result_offset,
                'limit': result_limit,
            }

            # Next range
            next_offset = result_offset + result_limit
            prev_offset = result_offset - result_limit
            if next_offset < item_count:
                original_url = self.request.build_absolute_uri()
                new_url = replace_querystring(original_url, offset=next_offset)
                meta_info['pagination']['next_url'] = new_url
            if prev_offset >= 0:
                original_url = self.request.build_absolute_uri()
                new_url = replace_querystring(original_url, offset=prev_offset)
                meta_info['pagination']['prev_url'] = new_url

        collection = map(self.model_presenter(*args, **kwargs), queryset)

        setattr(self, '_collection', collection)
        setattr(self, '_meta_info', meta_info)
        setattr(self, '_queryset', queryset)

        return collection, queryset, meta_info

    #-------------------------------------------------------------------------------------------------------------------
    # HTTP method

    def get(self, *args, **kwargs):
        collection, queryset, meta_info = self.get_collection(*args, **kwargs)


        if self.show_meta_in_response(*args, **kwargs):
            result = {
                'data': collection,
                'meta': meta_info,
            }
        else:
            result = collection
            self.response['X-Meta'] = json.dumps(meta_info)

        self.response.content = json.dumps(result)

    def post(self, *args, **kwargs):
        new_model = self.create_model(self.MODEL_COLLECTION_CLASS, *args, **kwargs)
        if new_model is None:
            raise NotImplementedError, 'create_model method doesn\'t return a new model object.'

        model_presentation = self.model_presenter(*args, **kwargs)(new_model)
        self.response.content = json.dumps(model_presentation)

    def delete(self, *args, **kwargs):
        self.get(*args, **kwargs)
        collection, queryset, meta_info = self.get_collection(*args, **kwargs)

        id_to_delete = [item['id'] for item in queryset.values('id')]
        items_to_delete = self.MODEL_COLLECTION_CLASS.objects.filter(id__in=id_to_delete)
        items_to_delete.delete()

    #-------------------------------------------------------------------------------------------------------------------
    # Public Method

    def show_meta_in_response(self, *args, **kwargs):
        return False

    def create_model(self, ModelClass, *args, **kwargs):
        return None

    def queryset(self, *args, **kwargs):
        ModelClass = self.MODEL_COLLECTION_CLASS
        if ModelClass is None:
            raise ImproperlyConfigured, 'Forget to set MODEL_CLASS for this ModelCollectionView'
        return ModelClass.objects.all()

    def order_list(self, query_set, orders, *args, **kwargs):
        result_order_list = []
        for direction, raw_field in orders:
            order = self.field_for_order(raw_field, direction)
            if order is None: continue
            if isinstance(order, (list, tuple)):
                result_order_list += order
            else:
                result_order_list.append(order)
        return  result_order_list

    def q_object_from_filter(self, field, filter_type, value, negate):
        q = Q(**{field+SUPPORTED_FILTER_TYPES_DICT[filter_type]: value})
        if negate: return ~q
        else: return q

    def field_for_order(self, order, direction, *args, **kwargs):
        return order if direction==1 else '-'+order

    def model_presenter(self, *args, **kwargs):
        return lambda x: getattr(x, 'json_dict')(summary=True)

#-----------------------------------------------------------------------------------------------------------------------

class ModelView(ModelViewBase):
    # Model
    MODEL_CLASS = None
    MODEL_NAME = None

    #-------------------------------------------------------------------------------------------------------------------
    # Object Lifecycle

    def __init__(self, *args, **kwargs):
        super(ModelView, self).__init__(*args, **kwargs)

        if self.MODEL_CLASS is None and kwargs.has_key('model_class'):
            self.MODEL_CLASS = kwargs['model_class']
        if self.MODEL_NAME is None:
            if kwargs.has_key('model_name'):
                self.MODEL_NAME = kwargs['model_name']
            else:
                self.MODEL_NAME = self.MODEL_CLASS.__name__

    #-------------------------------------------------------------------------------------------------------------------
    # HTTP Methods

    def get(self, *args, **kwargs):
        model_object = self.model_object(*args, **kwargs)
        model_presentation = self.model_presenter(*args, **kwargs)(model_object)

        self.response.content = json.dumps(model_presentation, separators=(',',':'))

    def put(self, *args, **kwargs):
        model_object = self.model_object(*args, **kwargs)

        new_model_object = self.update_model_object(model_object, *args, **kwargs)
        if new_model_object is not None:
            # Set cached one to updated one.
            self._model = new_model_object
        else:
            # Remove cached, let it update later
            delattr(self, '_model')

        self.get(*args, **kwargs) # response content

    def delete(self, *args, **kwargs):
        model_object = self.model_object(*args, **kwargs)
        self.get(*args, **kwargs) # response content
        model_object.delete()

    #-------------------------------------------------------------------------------------------------------------------
    # Public Methods

    def model_object(self, *args, **kwargs):
        # kwargs must contains "model_id" (from url)
        if hasattr(self, '_model'): return self._model

        ModelClass = kwargs['model_class'] if 'model_class' in kwargs else self.MODEL_CLASS
        if ModelClass is None:
            raise ImproperlyConfigured, 'Forget to set MODEL_CLASS for this ModelView'

        try:
            model_object = self.get_model(ModelClass, *args, **kwargs)
        except ModelClass.DoesNotExist:
            raise exceptions.ModelDoesNotExistError(self.MODEL_NAME)

        self._model = model_object
        setattr(self, self.MODEL_NAME.lower(), model_object)
        return model_object

    def update_model_object(self, original_model_object, *args, **kwargs):
        """For performance reason, it's good to return updated model object"""
        return original_model_object

    def get_model(self, ModelClass, model_id, *args, **kwargs):
        return ModelClass.objects.get(id=model_id)
